"""
Extra namespace for plugin tasks such as plugin.add
"""
import concurrent.futures
import datetime
import sys
import typing

import requests
from invoke import task, Context
from packaging.version import parse as parse_package_version, InvalidVersion, Version
from termcolor import colored
from ..meta import _pip, _get_available_plugins_from_pypi, _parse_versions, _gather_package_metadata_threaded


def _plugins(c: Context, pip_command=_pip()) -> list[str]:
    """
    List installed edwh-plugins
    """
    return c.run(f'{pip_command} freeze | grep edwh', hide=True, warn=True).stdout.strip().split("\n")


@task(name="list")
def list_plugins(c):
    """
        List installed plugins

        :param c: invoke ctx
        :type c: Context
        """
    available_plugins = ["edwh"] + _get_available_plugins_from_pypi('edwh', 'plugins')

    installed_plugins_raw = _plugins(c)

    if not installed_plugins_raw or len(installed_plugins_raw) == 1 and installed_plugins_raw[0] == "":
        raise ModuleNotFoundError("No 'edwh' packages found. That can't be right")

    installed_plugins = _parse_versions(installed_plugins_raw)

    plugin_info = _gather_package_metadata_threaded(available_plugins)

    old_plugins = []
    for plugin in available_plugins:
        metadata = plugin_info[plugin]
        current_version = installed_plugins.get(plugin, None)
        latest_version = parse_package_version(metadata["info"]["version"]) if metadata else None
        if latest_version and current_version:
            is_outdated = latest_version > current_version
        else:
            is_outdated = False

        github_url = metadata["info"]["project_urls"]["Documentation"]

        clean_name = plugin.removeprefix("edwh-").removesuffix("-plugin")
        if is_outdated:
            old_plugins.append(plugin)
            print(
                colored(
                    f"• {clean_name} ({latest_version} > {current_version}) - {github_url}",
                    'yellow',
                )
            )
        elif plugin in installed_plugins:
            print(
                colored(
                    f"• {clean_name} - {github_url}",
                    'green',
                )
            )
        else:
            print(
                colored(
                    f"◦ {clean_name} - {github_url}",
                    'red',
                )
            )

    if old_plugins:
        print()
        s = "" if len(old_plugins) == 1 else "s"
        print(colored(f"{len(old_plugins)} plugin{s} are out of date. Try `edwh self-update` to fix this.", "yellow"))


@task()
def add(c, plugin_name: str):
    """
    Install a new plugin

    Args:
        c (Context): invoke ctx
        plugin_name: which plugin to add
    """
    # ensure the prefix and suffix exist, but not twice:
    pip = _pip()

    plugin_name = plugin_name.removesuffix("-plugin").removeprefix("edwh-")
    plugin_name = f"edwh-{plugin_name}-plugin"

    c.run(f"{pip} install {plugin_name}")
